import { nativeDecimals, chainToPlatform, encoding, } from '@wormhole-foundation/connect-sdk';
import * as ethers_contracts from './ethers-contracts';
import { evmChainIdToNetworkChainPair } from './constants';
import { EvmAddress, EvmZeroAddress } from './address';
import { EvmPlatform } from './platform';
// forces EvmUtils to implement PlatformUtils
var _ = EvmUtils;
/**
 * @category EVM
 */
// Provides runtime concrete value
export var EvmUtils;
(function (EvmUtils) {
    function nativeTokenId(chain) {
        if (!isSupportedChain(chain))
            throw new Error(`invalid chain for EVM: ${chain}`);
        return {
            chain: chain,
            address: new EvmAddress(EvmZeroAddress), // TODO: fix weird type error
        };
    }
    EvmUtils.nativeTokenId = nativeTokenId;
    function isSupportedChain(chain) {
        const platform = chainToPlatform(chain);
        return platform === EvmPlatform.platform;
    }
    EvmUtils.isSupportedChain = isSupportedChain;
    function isNativeTokenId(chain, tokenId) {
        if (!isSupportedChain(chain))
            return false;
        if (tokenId.chain !== chain)
            return false;
        return tokenId.address.toString() === EvmZeroAddress;
    }
    EvmUtils.isNativeTokenId = isNativeTokenId;
    async function getDecimals(chain, rpc, token) {
        if (token === 'native')
            return nativeDecimals(EvmPlatform.platform);
        const tokenContract = getTokenImplementation(rpc, new EvmAddress(token).toString());
        return tokenContract.decimals();
    }
    EvmUtils.getDecimals = getDecimals;
    async function getBalance(chain, rpc, walletAddr, token) {
        if (token === 'native')
            return rpc.getBalance(walletAddr);
        const tokenImpl = getTokenImplementation(rpc, new EvmAddress(token).toString());
        return tokenImpl.balanceOf(walletAddr);
    }
    EvmUtils.getBalance = getBalance;
    async function getBalances(chain, rpc, walletAddr, tokens) {
        const balancesArr = await Promise.all(tokens.map(async (token) => {
            const balance = await getBalance(chain, rpc, walletAddr, token);
            const address = token === 'native' ? 'native' : new EvmAddress(token).toString();
            return { [address]: balance };
        }));
        return balancesArr.reduce((obj, item) => Object.assign(obj, item), {});
    }
    EvmUtils.getBalances = getBalances;
    async function sendWait(chain, rpc, stxns) {
        const txhashes = [];
        for (const stxn of stxns) {
            const txRes = await rpc.broadcastTransaction(stxn);
            txhashes.push(txRes.hash);
            if (chain === 'Celo') {
                console.error('TODO: override celo block fetching');
                continue;
            }
            // Wait for confirmation
            const txReceipt = await txRes.wait();
            if (txReceipt === null)
                continue; // TODO: throw error?
        }
        return txhashes;
    }
    EvmUtils.sendWait = sendWait;
    async function getCurrentBlock(rpc) {
        return await rpc.getBlockNumber();
    }
    EvmUtils.getCurrentBlock = getCurrentBlock;
    // Look up the Wormhole Canonical Network and Chain from the EVM chainId
    function chainFromChainId(eip155ChainId) {
        const ci = encoding.bignum.decode(eip155ChainId);
        const networkChainPair = evmChainIdToNetworkChainPair.get(ci);
        if (networkChainPair === undefined)
            throw new Error(`Unknown EVM chainId ${eip155ChainId}`);
        const [network, chain] = networkChainPair;
        return [network, chain];
    }
    EvmUtils.chainFromChainId = chainFromChainId;
    async function chainFromRpc(rpc) {
        const { chainId } = await rpc.getNetwork();
        return chainFromChainId(encoding.bignum.encode(chainId, true));
    }
    EvmUtils.chainFromRpc = chainFromRpc;
    function getTokenImplementation(connection, address) {
        const ti = ethers_contracts.TokenImplementation__factory.connect(address, connection);
        if (!ti)
            throw new Error(`No token implementation available for: ${address}`);
        return ti;
    }
    EvmUtils.getTokenImplementation = getTokenImplementation;
})(EvmUtils || (EvmUtils = {}));
//# sourceMappingURL=platformUtils.js.map